using System;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using XELF.Sphynx;

namespace XELF {

	public static class Helper {
		public static float ToRotation(this Vector2 normal) {
			var r = (float)Math.Atan2(normal.Y, normal.X);
			return r;
		}
		public static float Cross(Vector2 a, Vector2 b) {
			return a.X * b.Y - a.Y * b.X;
		}
		public static void Cross(ref Vector2 a, ref Vector2 b, out float result) {
			result = a.X * b.Y - a.Y * b.X;
		}
		public static void MAdd(ref Vector2 value, ref Vector2 addition, float scaleFactor, out Vector2 result) {
			Vector2 a;
			Vector2.Multiply(ref addition, scaleFactor, out a);
			Vector2.Add(ref value, ref a, out result);
		}
		public static bool SegmentSegmentIntersects(Vector2 segment0point0, Vector2 segment0point1,
								Vector2 segment1point0, Vector2 segment1point1,
								out Vector2 intersectionPoint,
								out float r, out float s) {

			Vector2 d0, d1;
			Vector2.Subtract(ref segment0point1, ref segment0point0, out d0);
			Vector2.Subtract(ref segment1point1, ref segment1point0, out d1);

			float d;
			Helper.Cross(ref d0, ref d1, out d);
			if (0 == d) {
				intersectionPoint = new Vector2(float.NaN, float.NaN);
				r = float.NaN;
				s = float.NaN;
				return false;
			}

			d = 1 / d;

			Vector2 vector;
			Vector2.Subtract(ref segment1point0, ref segment0point0, out vector);
			Helper.Cross(ref vector, ref d1, out r);
			Helper.Cross(ref vector, ref d0, out s);
			r *= d;
			s *= d;
			intersectionPoint = segment0point0 + r * d0;

			if (segment0point0.X == segment0point1.X) { intersectionPoint.X = segment0point0.X; }
			if (segment1point0.X == segment1point1.X) { intersectionPoint.X = segment1point0.X; }
			if (segment0point0.Y == segment0point1.Y) { intersectionPoint.Y = segment0point0.Y; }
			if (segment1point0.Y == segment1point1.Y) { intersectionPoint.Y = segment1point0.Y; }

			return r >= 0 && s >= 0 && r < 1 && s < 1;
		}
		public static void Curve(Vector3[] controlPoints, Vector3[] points, int splitsPerSegment) {
			int startIndex = 0;
			for (int i = 0, n = controlPoints.Length - 1; i < n; i++) {
				Curve(
					ref controlPoints[Math.Max(0, i - 1)],
					ref controlPoints[i],
					ref controlPoints[Math.Min(controlPoints.Length - 1, i + 1)],
					ref controlPoints[Math.Min(controlPoints.Length - 1, i + 2)],
					points, startIndex, splitsPerSegment);
				startIndex += splitsPerSegment;
			}
		}
		public static void Curve(ref Vector3 p0, ref Vector3 p1, ref Vector3 p2, ref Vector3 p3,
			Vector3[] points, int startIndex, int splits) {
			Vector3 start = p1, end;
			for (int i = 1; i <= splits; i++) {
				float amount = i * (1.0f / splits);
				Vector3.CatmullRom(ref p0, ref p1, ref p2, ref p3, amount, out end);
				points[startIndex + i - 1] = end;
				start = end;
			}
		}
	}

	public class CoasterGame : Microsoft.Xna.Framework.Game {
		GraphicsDeviceManager graphics;
		SpriteBatch spriteBatch;
		VectorGraphics vectorGraphics;
		SpriteFont font;
		Camera camera;
		Texture2D texture;

		float rotation;
		Vector2 gravity = new Vector2(0, 0.1f);
		Vector2 home = new Vector2(100, 100);

		int frame;
		float wait = 0;

		Vector2 accelaration;
		Vector2 velocity;
		Vector2 previousPosition;
		Vector2 position;
		SpriteEffects effects;
		bool inAir = true;
		int segment;
		readonly StringBuilder text = new StringBuilder();

		Vector2 min = new Vector2(-30, -30);
		Vector2 max = new Vector2(+30, +30);
		Vector2 offset = new Vector2(0x080, 0x100);

		Vector2 cameraPosition;
		Vector2 LeftAccelarationInAir = new Vector2(0.1f, 0);

		KeyboardState previousState;
		KeyboardState state;

		public CoasterGame() {
			graphics = new GraphicsDeviceManager(this);
			graphics.PreferMultiSampling = true;
			Content.RootDirectory = "Content";
		}

		protected override void LoadContent() {
			spriteBatch = new SpriteBatch(GraphicsDevice);
			font = Content.Load<SpriteFont>("DefaultFont");
			texture = Content.Load<Texture2D>("Run");
			vectorGraphics = new VectorGraphics(GraphicsDevice);
			camera = new Camera();
			vectorGraphics.Camera = camera;
			camera.Projection = Matrix.CreateOrthographicOffCenter(0, 800, 600, 0, -10, 10);
			position = previousPosition = home;

			points = new Vector3[splitsPerSegment * (controlPoints.Length - 1)];
			Helper.Curve(controlPoints, points, splitsPerSegment);
		}

		/// <summary>
		/// pȂiVւ̈ړAւ̒njB^̐ڐ(+/-)ɋB
		/// </summary>
		void Round() {
			var tangent = GetSegmentTangent();
			velocity = Vector2.Dot(velocity, tangent) * tangent;
			accelaration = Vector2.Dot(accelaration, tangent) * tangent;
			var start = GetSegmentStartPoint();
			Vector2 point;
			Vector2.Subtract(ref position, ref start, out point);
			position = Vector2.Dot(point, tangent) * tangent + start;
		}

		protected override void Update(GameTime gameTime) {
			if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
				this.Exit();

			if (inAir || Math.Abs(Vector2.Dot(GetSegmentTangent(), gravity)) >= gravity.Length() * 0.5f) {
				Vector2.Add(ref accelaration, ref gravity, out accelaration);
			}

			if (inAir) {
				Collides();
			} else {
				Round();
				rotation = GetSegmentTangent().ToRotation();
			}

			var speed = velocity.Length();
			wait -= speed * 0.1f;
			if (wait <= 0) {
				wait += 2;
				if (wait < 0) wait = 0;
				if (++frame >= 4) frame = 0;
			}
			if (speed < 0.1f) {
				frame = 2;
			}

			int direction;
			if (inAir) {
				direction = Math.Sign(Vector2.Dot(velocity, gravity));
			} else {
				direction = Math.Sign(Vector2.Dot(velocity, GetSegmentTangent()));
			}

			switch (direction) {
			case -1:
				effects = SpriteEffects.FlipHorizontally;
				break;
			case +1:
				effects = SpriteEffects.None;
				break;
			}

			previousState = state;
			state = Keyboard.GetState();

			accelaration *= 0.80f;
			velocity *= 0.97f;
			var a = 0.0f;
			if (state.IsKeyDown(Keys.A)) a += -0.1f;
			if (state.IsKeyDown(Keys.D)) a += +0.1f;
			if (state.IsKeyDown(Keys.Home)) position = previousPosition = home;
			if (state.IsKeyDown(Keys.W) && previousState.IsKeyUp(Keys.W) && !inAir) {
				Helper.MAdd(ref accelaration, ref gravity, -32.0f, out accelaration);
				Helper.MAdd(ref position, ref gravity, -32.0f, out position);
				inAir = true;
			}
			if (state.IsKeyDown(Keys.S) && previousState.IsKeyUp(Keys.S) && !inAir) {
				Helper.MAdd(ref accelaration, ref gravity, 1.0f, out accelaration);
				Helper.MAdd(ref position, ref gravity, 1.0f, out position);
				inAir = true;
			}

			if (!inAir) {
				var tangent = GetSegmentTangent();
				Helper.MAdd(ref accelaration, ref tangent, a, out accelaration);
			} else {
				Helper.MAdd(ref accelaration, ref LeftAccelarationInAir, a, out accelaration);
			}

			// ^
			previousPosition = position;
			Vector2.Clamp(ref velocity, ref min, ref max, out velocity);
			Vector2.Clamp(ref accelaration, ref min, ref max, out accelaration);
			Vector2.Add(ref velocity, ref accelaration, out velocity);
			Vector2.Add(ref position, ref velocity, out position);

			TrackSegment();

			text.Remove(0, text.Length);
			text.AppendFormat("InAir = {0}, Rotation = {1} [deg]", inAir, MathHelper.ToDegrees(rotation));
			if (!inAir) {
				text.AppendLine();
				text.AppendFormat("Segment {{ Index = {0}, Start = {1},\n Length = {2}, Tangent = {3} }}",
					segment, GetSegmentStartPoint(), GetSegmentLength(), GetSegmentTangent());
			}
			base.Update(gameTime);
		}

		// KvɉĐԂړB
		void TrackSegment() {
			int direction = 0;
			while (!inAir) {
				var p0 = GetSegmentStartPoint();
				var l = GetSegmentLength();
				var tangent = GetSegmentTangent();
				var t = (position - p0);
				float d;
				Vector2.Dot(ref tangent, ref t, out d);
				if (d < 0) {
					if (segment > 0) {
						if (direction > 0) return;
						direction = -1;
						segment--;
						Round();
					} else {
						inAir = true;
						return;
					}
				} else if (d > l) {
					if (segment < points.Length - 2) {
						if (direction < 0) return;
						direction = +1;
						segment++;
						Round();
					} else {
						inAir = true;
						return;
					}
				} else return;
			}
		}

		Vector2 GetSegmentStartPoint() {
			var d = points[segment];
			return new Vector2(d.X, d.Y);
		}
		Vector2 GetSegmentTangent() {
			var d = points[segment + 1] - points[segment];
			d.Normalize();
			return new Vector2(d.X, d.Y);
		}
		float GetSegmentLength() {
			var d = points[segment + 1] - points[segment];
			return d.Length();
		}

		Rectangle[] rectangles =
		{
			new Rectangle(0x100, 0x100, 0x100, 0x100), // 3
			new Rectangle(0x100, 0x000, 0x100, 0x100), // 1
			new Rectangle(0x000, 0x000, 0x100, 0x100), // 0
			new Rectangle(0x000, 0x100, 0x100, 0x100), // 2
		};
		const int splitsPerSegment = 16;
		Vector3[] points;
		Vector3[] controlPoints = new Vector3[] {
			new Vector3(300, 50, 0),
			new Vector3(0, 100, 0),
			new Vector3(-50, 60, 0),
			new Vector3(-50, 40, 0),
			new Vector3(0, 0, 0),
			new Vector3(200, 250, 0),
			new Vector3(300, 300, 0),
			new Vector3(400, 250, 0),
			new Vector3(410, 200, 0),
			new Vector3(380, 130, 0),
			new Vector3(330, 100, 0),
			new Vector3(280, 100, 0),
			new Vector3(200, 140, 0),
			new Vector3(200, 170, 0),
			new Vector3(220, 200, 0),
			new Vector3(250, 230, 0),
			new Vector3(400, 250, 0),
			new Vector3(500, 250, 0),
			new Vector3(800, 200, 0),
			new Vector3(1200, 260, 0),
			new Vector3(1250, 200, 0),
			new Vector3(1300, 260, 0),
			new Vector3(1600, 260, 0),
		};

		/// <summary>
		/// LN^[Ɛ̏Փ˂oāAՓ˂ĂΒnB
		/// </summary>
		void Collides() {
			Vector2 d;
			Vector2.Subtract(ref previousPosition, ref position, out d);
			if (d.LengthSquared() > 0) {
				float distanceSquared = float.PositiveInfinity;
				int closestSegment = -1;

				for (int i = 1; i < points.Length; i++) {
					Vector2 s0 = new Vector2(points[i - 1].X, points[i - 1].Y);
					Vector2 s1 = new Vector2(points[i].X, points[i].Y);
					Vector2 intersectionPoint;
					float r, s;
					if (Helper.SegmentSegmentIntersects(previousPosition, position, s0, s1,
						out intersectionPoint, out r, out s)) {
						float dd;
						Vector2.DistanceSquared(ref intersectionPoint, ref previousPosition, out dd);
						if (dd < distanceSquared) {
							distanceSquared = dd;
							closestSegment = i;
						}
					}
				}

				if (closestSegment >= 0) {
					segment = closestSegment - 1;
					inAir = false;
					Round();
				}
			}
		}

		protected override void Draw(GameTime gameTime) {
			GraphicsDevice.Clear(Color.SteelBlue);

			cameraPosition = position - new Vector2(800 * 0.5f, 600 * 0.5f);

			vectorGraphics.Camera.View = Matrix.CreateTranslation(-cameraPosition.X, -cameraPosition.Y, 0);
			vectorGraphics.Begin();
			vectorGraphics.DrawLine(points, Color.White);
			vectorGraphics.End();

			spriteBatch.Begin();
			spriteBatch.Draw(texture, position - cameraPosition, rectangles[frame], Color.White, rotation, offset, 0.5f, effects, 0);
			spriteBatch.DrawString(font, text, Vector2.Zero, Color.White);
			spriteBatch.End();

			base.Draw(gameTime);
		}
	}
}
